from multiprocessing import Process
from csv import writer as csv_writer
from argparse import ArgumentParser
from time import sleep
import shutil
import os
from functools import lru_cache

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.cm import get_cmap
from scipy.interpolate import interp1d
from scipy.spatial.distance import cdist
import numpy as np
import cv2
from rich import print
from rich.console import Console
from rich.table import Table

from id_manual_tools.utils import file_path, trajectory_path
from id_manual_tools.set_corners import arg_main as set_corners
from id_manual_tools.matplotlib_gui import matplotlib_gui
from id_manual_tools.get_nans import get_list_of_nans_from_traj

# from PyQt5.QtWidgets import QToolBar
console = Console()


class trajectory_corrector(matplotlib_gui):
    def __init__(
        self,
        video_path,
        traj_path,
        setup_points=None,
        jumps_check_sigma=None,
        automatic_check=-1,
        fps=None,
        n_cores=4,
    ):
        self.jumps_check_sigma = jumps_check_sigma
        console.rule("[green]Welcome to the id_manual_tools manual validator")
        self.automatic_check = automatic_check
        self.video_path = os.path.abspath(video_path)
        self.traj_path = os.path.abspath(traj_path)

        self.preloaded_frames_path = os.path.abspath("Preloaded_frames")
        video_path_file = os.path.join(self.preloaded_frames_path, "video_path.txt")

        os.makedirs(self.preloaded_frames_path, exist_ok=True)
        try:
            with open(video_path_file, "r") as file:
                if file.readline() != self.video_path:
                    raise FileNotFoundError
        except FileNotFoundError:
            print(
                f"Creating new preloaded frames directory: {self.preloaded_frames_path}"
            )
            shutil.rmtree(self.preloaded_frames_path)
            os.makedirs(self.preloaded_frames_path)
            with open(video_path_file, "w") as file:
                file.write(self.video_path)
        else:
            print(f"Reusing frames from {self.preloaded_frames_path}")

        # TODO ask for an exclusion area with set_corners and set all points inside as nans
        self.cap = cv2.VideoCapture(video_path)
        print(f"Loaded video {self.video_path}")

        self.data = np.load(self.traj_path, allow_pickle=True).item()
        print(f"Loaded data  {self.traj_path}")

        if setup_points is not None:
            try:
                exist_required_setup_points = setup_points in self.data["setup_points"]
            except (KeyError, TypeError):
                exist_required_setup_points = False
            while not exist_required_setup_points:
                print(
                    f"Setup_points setter is launched to define the user required setup_points {setup_points}"
                )
                set_corners(self.video_path, self.data)

                try:
                    exist_required_setup_points = (
                        setup_points in self.data["setup_points"]
                    )
                except (KeyError, TypeError):
                    exist_required_setup_points = False

            corners = self.data["setup_points"][setup_points]
            self.xmin = int(np.min(corners[:, 0]))
            self.xmax = int(np.max(corners[:, 0]))
            self.ymin = int(np.min(corners[:, 1]))
            self.ymax = int(np.max(corners[:, 1]))
        else:
            self.xmin = 0
            self.xmax = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
            self.ymin = 0
            self.ymax = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

        if fps:
            if fps != self.data["frames_per_second"]:
                self.data["frames_per_second"] = fps
                print(f"Frames per second updated to {fps}")

        self.all_traj = self.data["trajectories"]
        self.total_frames, self.N = self.all_traj.shape[:2]
        assert self.total_frames == self.cap.get(cv2.CAP_PROP_FRAME_COUNT)

        self.N -= 1
        self.BL = self.data["body_length"]

        self.not_preloaded_frame = np.empty(self.total_frames, bool)
        for frame in range(self.total_frames):
            self.not_preloaded_frame[frame] = not os.path.exists(
                os.path.join(self.preloaded_frames_path, f"{frame}.npz")
            )

        copy_of_traj = np.zeros_like(self.all_traj)
        if jumps_check_sigma is not None:
            vel = np.linalg.norm(np.diff(self.all_traj, axis=0), axis=2)
            impossible_jumps = vel > (
                np.nanmean(vel) + jumps_check_sigma * np.nanstd(vel)
            )
            copy_of_traj[:-1][impossible_jumps] = np.nan
            print(f"Number of impossible jumps: {np.sum(impossible_jumps)}")
            self.list_of_jumps = get_list_of_nans_from_traj(
                copy_of_traj, sort_by="start"
            )
        else:
            self.list_of_jumps = []

        self.list_of_nans = get_list_of_nans_from_traj(self.all_traj, sort_by="start")

        self.write_lists_of_nans_and_jumps()

        self.limits = (self.xmin, self.xmax, self.ymin, self.ymax)
        print(f"xmin, xmax, ymin, ymax = {self.limits}")
        self.Lx = 0.5 * (self.xmax - self.xmin)
        self.Ly = 0.5 * (self.ymax - self.ymin)
        self.pad = 7
        self.pad_extra = 150
        self.actual_plotted_frame = -1
        self.G_pressed = False

        if self.list_of_nans or self.list_of_jumps:
            list_of_frames_to_preload = set()
            for id, start, end, duration in self.list_of_nans + self.list_of_jumps:
                if duration <= self.automatic_check:
                    list_of_frames_to_preload.add(max(0, start - 1))
                else:
                    pad = min(self.pad, duration)
                    for frame in range(
                        max(0, start - pad), min(self.total_frames, end + pad)
                    ):
                        list_of_frames_to_preload.add(frame)
            list_of_frames_to_preload = list(list_of_frames_to_preload)
            print(f"{len(list_of_frames_to_preload)} frames needed")
            list_of_frames_to_preload = [
                frame
                for frame in list_of_frames_to_preload
                if not os.path.exists(
                    os.path.join(self.preloaded_frames_path, f"{frame}.npz")
                )
            ]
            print(f"{len(list_of_frames_to_preload)} frames to preload")
            list_of_frames_to_preload.sort()
            if list_of_frames_to_preload:
                self.preload_frames_list(list_of_frames_to_preload, n_cores=n_cores)

            self.create_figure()

            self.Delta = 1
            if self.list_of_nans:
                self.next_episode(self.list_of_nans.pop(-1))
            elif self.list_of_jumps:
                self.next_episode(self.list_of_jumps.pop(-1))

            plt.show()
        else:
            if jumps_check_sigma is not None:
                print("[red]There's no nans nor impossible jumps to correct")
            else:
                print("[red]There's no nans to correct")

    def next_episode(self, params):
        self.id, self.start, self.end, _ = params

        console.rule(
            f"[bold red]Episode for fish {self.id} from {self.start} to {self.end}, {self.end-self.start} nans"
        )
        self.id_traj = self.all_traj[:, self.id, :]
        self.traj = np.delete(self.all_traj, self.id, axis=1)

        if self.N:
            temp = self.traj.reshape(-1, self.N, 1, 2)
            self.segments = np.concatenate([temp[:-1], temp[1:]], axis=2)

        self.frame = max(0, self.start - 1)

        self.zoom = 0.3

        if not np.isnan(self.id_traj[self.frame, 0]):
            self.x_center, self.y_center = self.id_traj[self.frame]
        else:
            self.x_center, self.y_center = np.nanmean(self.traj[self.frame], axis=0)
        self.set_ax_lims(do_not_draw=True)
        self.interpolation_range = np.arange(self.start, self.end)
        self.continuous_interpolation_range = np.arange(self.start - 1, self.end, 0.2)

        self.user_detection_history = []

        self.fit_interpolator_and_draw_frame()

        if (self.end - self.start) <= self.automatic_check:
            sleep(0.1)
            self.key_enter()

    def preload_frames_list(self, list_of_frames, n_cores=10):

        n_frames = len(list_of_frames)
        chunks = max(50, n_frames // n_cores)
        print(
            f"[red]Starting {len(range(0, n_frames, chunks))} processes of {chunks} frames each"
        )

        # self.process_frame_list_and_save(
        #     self.video_path,
        #     list_of_frames,
        #     self.process_image,
        #     self.limits,
        # )
        for s in range(0, n_frames, chunks):
            # self.not_preloaded_frame[start:end] = False
            Process(
                target=trajectory_corrector.process_frame_list_and_save,
                args=(
                    self.preloaded_frames_path,
                    self.video_path,
                    list_of_frames[s : s + chunks],
                    trajectory_corrector.process_image,
                    self.limits,
                ),
            ).start()

    def draw_frame(self):

        self.points.set_offsets(self.traj[self.frame])
        if self.frame in self.interpolation_range:
            self.id_point.set_offsets(self.interpolator(self.frame))
        else:
            self.id_point.set_offsets(self.id_traj[self.frame])
        self.interpolated_points.set_data(self.interpolator(self.interpolation_range))
        self.interpolated_line.set_data(
            self.interpolator(self.continuous_interpolation_range)
        )

        self.interpolated_train.set_data(*self.interpolator.y)

        if self.frame != self.actual_plotted_frame:
            self.im.set_data(self.get_frame(self.frame))

            # self.im._A = self.get_frame(self.frame)
            # self.im._imcache = None
            # self.im._rgbacache = None
            # self.im.stale = True

            self.text.set_text(f"Frame {self.frame}")
            self.actual_plotted_frame = self.frame

        origin = max(0, self.frame - 30)
        for fish in range(self.N):
            self.LineCollections[fish].set_segments(
                self.segments[origin : self.frame, fish]
            )
        self.draw_and_flush()

    def create_figure(self):
        super().__init__("Trajectory correction")

        (self.interpolated_line,) = self.ax.plot([], [], "w-", zorder=8)
        (self.interpolated_points,) = self.ax.plot([], [], "w.", zorder=8)
        (self.interpolated_train,) = self.ax.plot([], [], "r.", zorder=9)

        self.im = self.ax.imshow(
            [[]],
            cmap="gray",
            vmax=255,
            vmin=0,
            extent=(
                self.xmin,
                self.xmax,
                self.ymax,
                self.ymin,
            ),
            interpolation="none",
            animated=True,
            resample=False,
            snap=False,
        )

        cmap = get_cmap("gist_rainbow")
        self.points = self.ax.scatter(
            *np.zeros((2, self.N)),
            c=cmap(np.arange(self.N) / (self.N - 1)),
            s=10.0,
        )

        self.id_point = self.ax.scatter([], [], c="k", s=10.0, zorder=10)
        self.text = self.ax.text(
            0.1, 0.1, "", size=15, zorder=15, transform=self.ax.transAxes
        )

        line_lenght = 30
        self.LineCollections = []
        for i in range(self.N):
            color = np.tile(cmap(i / (max(1, self.N - 1))), (line_lenght, 1))
            color[:, -1] = np.linspace(0, 1, line_lenght)
            self.LineCollections.append(LineCollection([], linewidths=2, color=color))

        for linecollection in self.LineCollections:
            self.ax.add_collection(linecollection)

    def fit_interpolator_and_draw_frame(self):

        time_range = np.arange(
            max(0, self.start - (self.pad + self.pad_extra)),
            min(self.total_frames, self.end + (self.pad + self.pad_extra)),
        )

        time_range = time_range[~np.isnan(self.id_traj[time_range, 0])]

        self.interpolator = interp1d(
            time_range,
            self.id_traj[time_range].T,
            axis=1,
            kind="cubic",
            fill_value="extrapolate",
        )
        self.draw_frame()

    def key_a(self, draw=True):
        """Go back Delta timesteps"""
        self.frame = max(0, self.frame - self.Delta)
        if draw:
            self.draw_frame()

    def key_d(self, draw=True):
        """Advance Delta timesteps"""
        self.frame = min(self.total_frames - 1, self.frame + self.Delta)
        if draw:
            self.draw_frame()

    def key_left(self):
        """Go back Delta timesteps"""
        self.key_a()

    def key_right(self):
        """Advance Delta timesteps"""
        self.key_d()

    def key_P(self):
        """Toggle 1500 extra timesteps in the interpolator data"""
        if self.pad_extra == 1500:
            self.pad_extra = 0
        else:
            self.pad_extra = 1500

        self.fit_interpolator_and_draw_frame()

    def key_p(self):
        """Toggle 150 extra timesteps in the interpolator data"""
        if self.pad_extra == 150:
            self.pad_extra = 0
        else:
            self.pad_extra = 150

        self.fit_interpolator_and_draw_frame()

    def key_z(self):
        """Undo the last point defined by user in the interpolation range"""
        if self.user_detection_history:
            frame, position = self.user_detection_history.pop()
            self.id_traj[frame] = position

            self.fit_interpolator_and_draw_frame()

    def key_e(self):
        """Togle actual frame from end to start and recenter"""
        if self.frame == self.end:
            self.frame = max(0, self.start - 1)
        else:
            self.frame = self.end
        self.recenter()
        self.draw_frame()

    def key_n(self):
        """Sets the actual position to nan (only on the boundaries and inside of the interpolation range)"""
        if self.frame == (self.start - 1) or self.frame == self.end:

            if self.frame == (self.start - 1):
                self.id_traj[
                    max(0, self.frame - self.Delta + 1) : self.frame + 1
                ] = np.nan
                while np.isnan(self.id_traj[self.frame, 0]):
                    self.start -= 1
                    self.frame -= 1
                    if self.start == 0:
                        self.frame = 0
                        break
            elif self.frame == self.end:
                self.id_traj[
                    self.frame : min(self.total_frames, self.frame + self.Delta)
                ] = np.nan
                while np.isnan(self.id_traj[self.frame, 0]):
                    self.end += 1
                    self.frame += 1
                    if self.end == (self.total_frames - 1):
                        break

            self.interpolation_range = np.arange(self.start, self.end)
            self.continuous_interpolation_range = np.arange(
                self.start - 1, self.end + 0.1, 0.2
            )
            self.fit_interpolator_and_draw_frame()

        elif (self.end - self.frame) < self.Delta:

            self.id_traj[
                self.frame : min(self.total_frames, self.frame + self.Delta)
            ] = np.nan

            while np.isnan(self.id_traj[self.frame, 0]):
                self.frame += 1
                if self.frame == (self.total_frames - 1):
                    break
            self.end = self.frame

            self.fit_interpolator_and_draw_frame()
        elif self.frame in self.interpolation_range:
            self.id_traj[self.frame] = np.nan
            self.fit_interpolator_and_draw_frame()

        else:
            print(f"You are not on the boundaries, you are at frame {self.frame}")
            print(
                f"You only can set nan values on frames {self.start-1} and {self.end}"
            )

    def key_enter(self):
        """Accept the interpolation, write it to the trajectory array and move on (this doesn't write on disk)"""
        print(
            f"Writting interploation into the array from {self.start} to {self.end} for fish {self.id}"
        )
        self.id_traj[self.interpolation_range] = self.interpolator(
            self.interpolation_range
        ).T

        self.list_of_nans = get_list_of_nans_from_traj(self.all_traj, sort_by="start")

        if self.list_of_nans:
            self.next_episode(self.list_of_nans.pop(-1))
        elif self.list_of_jumps:
            self.next_episode(self.list_of_jumps.pop(-1))
        else:
            self.key_w()
            plt.close()

    def key_w(self):
        """Write on disk the actual state of the trajectory array"""
        print(f"Saving data to {self.traj_path}")
        np.save(self.traj_path, self.data)
        self.write_lists_of_nans_and_jumps()

    def write_lists_of_nans_and_jumps(self):
        with open("list_of_nans.csv", "w", newline="") as csvfile:
            csvfile.write("fish_id,start,end,duration\n")
            writer = csv_writer(csvfile)
            writer.writerows(self.list_of_nans)
        print(f"List of nans saved at {os.path.abspath('list_of_nans.csv')}")

        if self.jumps_check_sigma is not None:
            with open("list_of_jumps.csv", "w", newline="") as csvfile:
                csvfile.write("fish_id,start,end,duration\n")
                writer = csv_writer(csvfile)
                writer.writerows(self.list_of_jumps)
            print(f"List of jumps saved at {os.path.abspath('list_of_jumps.csv')}")

    def recenter(self):
        if self.frame in self.interpolation_range:
            self.x_center, self.y_center = self.interpolator(self.frame)
        else:
            self.x_center, self.y_center = self.id_traj[self.frame]
        self.set_ax_lims(do_not_draw=True)

    def key_g(self):
        """Apply key d and key x sequentially"""
        self.key_d(draw=False)
        self.key_x(recenter=True)

    def key_G(self):
        """Apply key g until the end of the episode is reached"""
        self.G_pressed = not self.G_pressed
        while self.G_pressed:
            self.key_g()
            if self.frame >= self.end:
                self.G_pressed = False
        self.draw_frame()

    def key_number(self, number):
        if number:
            self.Delta = 2 ** (number - 1)

    def key_h(self):
        """Shows Key Bindings table"""
        table = Table(title="Key Bindings")

        keys = [
            "d",
            "right",
            "a",
            "left",
            "p",
            "P",
            "n",
            "z",
            "x",
            "g",
            "enter",
            "w",
            "h",
        ]
        table.add_column("Key", justify="center", style="cyan", no_wrap=True)
        table.add_column("Description", justify="center", style="magenta")

        for key in keys:
            table.add_row(key, getattr(self, f"key_{key}").__doc__)

        table.add_row("1-9", "Set Delta to 1, 2, 4, 8, 16, 32, 64, 128, 256")
        table.add_row("s", "Save screenshot")
        table.add_row("f", "Toggle full screen")
        table.add_row("q", "Quit application")
        console.print(table)

    def key_x(self, recenter=False):
        """Set the actual position of the blob by finding the center of mass of the image drawned around the interpolated position"""
        if recenter:
            self.recenter()
        if self.frame in self.interpolation_range:
            self.find_blob(*self.interpolator(self.frame))

    @lru_cache(maxsize=1024)
    def get_frame(self, frame):
        path = os.path.join(self.preloaded_frames_path, f"{frame}.npz")
        if os.path.exists(path):
            return np.load(path)["arr_0"]

        print(f"[red]Had to load frame {frame}")
        if self.cap.get(cv2.CAP_PROP_POS_FRAMES) != frame:
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
        ret, image = self.cap.read()
        assert ret

        image = self.process_image(image, *self.limits)
        np.save(path, image)
        return image
        # return np.ma.masked_invalid(
        #     self.process_image(image, *self.limits)
        # ).shrink_mask()

    @staticmethod
    def process_image(image, xmin, xmax, ymin, ymax):
        image = np.mean(image[ymin:ymax, xmin:xmax], axis=-1)
        image -= np.min(image)
        image *= 255 / np.max(image)
        image = np.uint8(image)
        return image

    def find_blob(self, x, y):
        if x <= self.xmin or x >= self.xmax or y >= self.ymax or y <= self.ymin:
            return

        canvas_x_min = max(0, int(x - self.BL - self.xmin))
        canvas_y_min = max(0, int(y - self.BL - self.ymin))

        canvas_center = (x - canvas_x_min - self.xmin, y - canvas_y_min - self.ymin)

        fish_im = self.get_frame(self.frame)[
            canvas_y_min : int(y + self.BL - self.ymin),
            canvas_x_min : int(x + self.BL - self.xmin),
        ]

        fish_im = cv2.GaussianBlur(fish_im, (0, 0), 2)

        _, mask = cv2.threshold(
            fish_im,
            0,
            1,
            cv2.THRESH_BINARY + cv2.THRESH_OTSU,
        )

        contours, hierarchy = cv2.findContours(
            1 - mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
        )

        blobs_positions = []
        for c in contours:
            M = cv2.moments(c)
            blobs_positions.append((M["m10"] / M["m00"], M["m01"] / M["m00"]))

        closer_blob = cdist(
            [
                canvas_center,
            ],
            blobs_positions,
        ).argmin()

        x_c, y_c = blobs_positions[closer_blob]

        # fig, ax = plt.subplots()
        # ax.imshow(fish_im, cmap="gray")
        # ax.plot(*blobs_positions.T, "r.")
        # ax.plot(*canvas_center, "b.")
        # ax.plot(x_c, y_c, "g.")
        # fig.savefig("id_dev/fish", dpi=300)

        self.user_detection_history.append(
            (self.frame, tuple(self.id_traj[self.frame]))
        )
        self.id_traj[self.frame] = (
            x_c + canvas_x_min + self.xmin,
            y_c + canvas_y_min + self.ymin,
        )
        self.fit_interpolator_and_draw_frame()

    def button_3(self, event):
        self.find_blob(event.xdata, event.ydata)

    def button_1(self, event):
        self.user_detection_history.append(
            (self.frame, tuple(self.id_traj[self.frame]))
        )
        self.id_traj[self.frame] = event.xdata, event.ydata
        self.fit_interpolator_and_draw_frame()

    def on_key(self, event):
        try:
            int_key = int(event.key)
            if int_key in range(1, 10):
                self.Delta = 2 ** (int_key - 1)
        except ValueError:
            try:
                fun = getattr(self, f"key_{event.key}")
                fun()
            except AttributeError:
                pass

    @staticmethod
    def process_frame_list_and_save(
        save_dir, video_path, list_of_frames, process_fun, lims
    ):
        # cv2.setNumThreads(1)
        cap = cv2.VideoCapture(video_path)
        for frame in list_of_frames:
            if cap.get(cv2.CAP_PROP_POS_FRAMES) != frame:
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
            ret, image = cap.read()
            assert ret
            np.savez_compressed(
                os.path.join(save_dir, f"{frame}"),
                process_fun(image, *lims),
            )
        print(
            f"Preloaded episode with frames {list_of_frames[0]} => {list_of_frames[-1]}"
        )


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "-s",
        metavar="session",
        type=str,
        help="idTracker.ai succesful session directory or trajectory file",
        required=True,
    )
    parser.add_argument(
        "-video",
        type=file_path,
        help="Video file (only one file)",
        required=True,
    )

    parser.add_argument(
        "-jumps_check_sigma",
        type=float,
        default=None,
        help="Check for impossible long jumps on the trajectories",
    )

    parser.add_argument(
        "-reset",
        action="store_true",
        default=False,
        help="Ignores any previously edited file",
    )

    parser.add_argument(
        "-auto_validation",
        default=0,
        type=int,
        help="Max length of nan episode to apply auto-correction",
    )

    parser.add_argument(
        "-fps",
        default=0,
        type=int,
        help="Overwrite the frame rate of the session",
    )
    parser.add_argument(
        "-n",
        type=int,
        default=10,
        help="number of threads for parallel processing. Default is 4",
    )

    args = parser.parse_args()

    trajectory_corrector(
        args.video,
        trajectory_path(args.s, reset=args.reset),
        jumps_check_sigma=args.jumps_check_sigma,
        automatic_check=args.auto_validation,
        setup_points="corners_out",
        fps=args.fps,
        n_cores=args.n,
    )


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt


class matplotlib_gui:
    """Thats a base class for building ultra simple matplotlib aplications with background images. Used in get_corners and correct_trajectories"""

    def draw_and_flush(self):
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def __init__(self, title=" "):

        self.zoom = 1
        self.Ly = 500  # px
        self.Lx = 500  # px
        self.x_center = 0
        self.y_center = 0
        self.mouse_pressed = False
        self.has_moved = False

        self.fig = plt.figure(figsize=(8, 8))
        self.ax = self.fig.add_axes(
            [0, 0, 1, 1],
            xticks=(),
            yticks=(),
            facecolor="gray",
        )

        self.canvas_size = self.fig.get_size_inches() * self.fig.dpi

        # self.fig.canvas.manager.window.findChild(QToolBar).setVisible(False)
        self.fig.canvas.manager.set_window_title(title)
        self.fig.canvas.mpl_connect("button_press_event", self.on_click)
        self.fig.canvas.mpl_connect("button_release_event", self.on_click_release)
        self.fig.canvas.mpl_connect("key_release_event", self.on_key)
        self.fig.canvas.mpl_connect("scroll_event", self.on_scroll)
        self.fig.canvas.mpl_connect("motion_notify_event", self.on_motion)
        self.fig.canvas.mpl_connect("resize_event", self.on_resize)

    def on_click(self, event):
        self.has_moved = False
        self.mouse_pressed = True
        self.click_origin = (event.x, event.y)

    def on_click_release(self, event):
        self.mouse_pressed = False
        if not self.has_moved:
            if hasattr(self, f"button_{event.button}"):
                getattr(self, f"button_{event.button}")(event)

    def on_key(self, event):
        try:
            int_key = int(event.key)
        except ValueError:
            try:
                getattr(self, f"key_{event.key}")()
            except AttributeError:
                pass
        else:
            if hasattr(self, "key_number"):
                self.key_number(int_key)

    def on_scroll(self, event):
        self.x_center += (self.x_center - event.xdata) * 0.1 * event.step
        self.y_center += (self.y_center - event.ydata) * 0.1 * event.step
        self.zoom += 0.1 * self.zoom * event.step
        self.set_ax_lims()

    def on_motion(self, event):
        if self.mouse_pressed:
            self.has_moved = True
            self.x_center -= (
                2
                * self.zoom
                * self.Lx
                * (event.x - self.click_origin[0])
                / self.canvas_size[0]
            )
            self.y_center += (
                2
                * self.zoom
                * self.Ly
                * (event.y - self.click_origin[1])
                / self.canvas_size[1]
            )
            self.click_origin = (event.x, event.y)
            self.set_ax_lims()

    def on_resize(self, event):
        self.Ly = event.height * self.Ly / self.canvas_size[1]
        self.Lx = event.width * self.Lx / self.canvas_size[0]
        self.canvas_size = (event.width, event.height)
        self.set_ax_lims()

    def set_ax_lims(self, do_not_draw=False):
        self.ax.set(
            xlim=(
                self.x_center - self.zoom * self.Lx,
                self.x_center + self.zoom * self.Lx,
            ),
            ylim=(
                self.y_center + self.zoom * self.Ly,
                self.y_center - self.zoom * self.Ly,
            ),
        )
        if not do_not_draw:
            self.fig.canvas.draw()



from matplotlib.pyplot import Figure
import numpy as np
import cv2
from matplotlib.animation import FuncAnimation
from scipy.signal import savgol_filter
from matplotlib.collections import LineCollection
from matplotlib.cm import get_cmap
from argparse import ArgumentParser
from tqdm import trange
from multiprocessing import Pool
from subprocess import call
from os import makedirs, remove
from os.path import splitext
from shutil import rmtree
from warnings import catch_warnings, simplefilter
from scipy.ndimage import gaussian_filter1d


def interpolate_nans(arr):
    isnan = np.isnan(arr)
    arr[isnan] = np.interp(np.where(isnan)[0], xp=np.where(~isnan)[0], fp=arr[~isnan])


cmap = get_cmap("gist_rainbow")


# INPUT ARGUMENTS (ARGPARSER)
parser = ArgumentParser()
parser.add_argument(
    "-i", metavar="video", type=str, help="Input raw video file", required=True
)
parser.add_argument(
    "-traj",
    type=str,
    help="Path to trajectories to load (.npy file)",
    required=True,
)
parser.add_argument(
    "-o",
    metavar="output",
    type=str,
    help="Output file name, default is (input)_tracked.mp4",
    default=None,
)
parser.add_argument(
    "-t",
    metavar="time",
    type=float,
    help="Duration of the tracked video (in seconds), default is entire video",
    default=None,
)
parser.add_argument(
    "-z",
    action="store_true",
    default=False,
    help="Activates dynamic zoom in the video",
)

parser.add_argument(
    "-vmin",
    default=0,
    help="Minimum value for video colormap, default 0",
)

parser.add_argument(
    "-vmax",
    default=255,
    help="Maximum value for video colormap, default 255",
)
parser.add_argument("-n", type=int, default=4, help="number of threads. Default is 4")

args = parser.parse_args()
line_lenght = 20
tmp_folder = "/tmp/plot_trajectories_tmp/"


# LOADING DATA (POSITIONS AND VIDEO)
data = np.load(args.traj, allow_pickle=True).item()
pos = data["trajectories"]
n_frames, N, _ = pos.shape

cap = cv2.VideoCapture(args.i)
fps = cap.get(cv2.CAP_PROP_FPS)
nx = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
ny = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
cap.release
cap = None


corners = data["setup_points"]["corners_out"]
xmin = int(np.min(corners[:, 0]))
xmax = int(np.max(corners[:, 0]))
ymin = int(np.min(corners[:, 1]))
ymax = int(np.max(corners[:, 1]))

#  COMPUTING DYNAMIC ZOOM (CENTER OF MASS AND ZOOM RADI)
if args.z:
    with catch_warnings():
        simplefilter("ignore", category=RuntimeWarning)
        c_mass = np.nanmean(pos, axis=1)
        z_radi = np.sqrt(
            np.mean(np.nansum((pos - c_mass[:, None, :]) ** 2, axis=-1), axis=-1)
        )

    # interpolate_nans(c_mass[:, 0])
    # interpolate_nans(c_mass[:, 1])
    # interpolate_nans(z_radi)

    c_mass = savgol_filter(c_mass, 31, 3, axis=0, mode="nearest")
    z_radi = savgol_filter(z_radi, 51, 3, mode="nearest")

    zoom = 2

    x_min = np.clip(c_mass[:, 0] - zoom * z_radi, xmin, None)
    y_min = np.clip(c_mass[:, 1] - zoom * z_radi, ymin, None)
    x_max = np.clip(x_min + 2 * zoom * z_radi, None, xmax)
    y_max = np.clip(y_min + 2 * zoom * z_radi, None, ymax)

    x_min = gaussian_filter1d(x_min, sigma=10)
    y_min = gaussian_filter1d(y_min, sigma=10)
    x_max = gaussian_filter1d(x_max, sigma=10)
    y_max = gaussian_filter1d(y_max, sigma=10)

    c_mass = None
    z_radi = None


# PREPARING PLOT LINES "SEGMENTS"
temp = pos.reshape(-1, N, 1, 2)
segments = np.concatenate([temp[:-1], temp[1:]], axis=2)
temp = None


# DISTRIBUTE FRAMES TO PROCESS BETWEEN THREADS
output_frames = int(args.t * fps) if args.t else n_frames
frame_limits = np.linspace(0, output_frames, args.n + 1, dtype=int)
frame_limits = np.column_stack([frame_limits[:-1], frame_limits[1:]])


# INITIALIZING FIGURE
dpi = 100
if args.z:
    fig = Figure(figsize=(15, 15), dpi=dpi)
else:
    Lx = xmax - xmin
    Ly = ymax - ymin
    diag = np.sqrt(Lx * Lx + Ly * Ly) / np.sqrt(2)
    fig = Figure(figsize=(15 * Lx / diag, 15 * Ly / diag), dpi=dpi)

fig.patch.set_facecolor("black")
axs = fig.add_axes([0, 0, 1, 1], facecolor="k", xticks=(), yticks=())
axs.set(xticks=(), yticks=(), xlim=(xmin, xmax), ylim=(ymax, ymin))

# colors = cmap((pos[0, :, 0] > 2000).astype(float) * 0.5)
colors = cmap(np.arange(N) / N)

points = axs.scatter(*np.zeros((2, N)), c=colors, s=20.0)

im = axs.imshow(
    np.zeros((ny, nx)), origin="upper", cmap="gray", vmax=args.vmax, vmin=args.vmin
)

LineCollections = []  # DON'T ASK...
for i in range(N):
    color = np.tile(colors[i], (line_lenght, 1))
    color[:, -1] = np.linspace(0, 1, line_lenght)
    LineCollections.append(axs.add_collection(LineCollection([], color=color)))


def thread_run(thread):
    """
    This function runs in every thread generating a portion of the final video
    """
    start, finish = frame_limits[thread]
    n_frames = finish - start

    cap = cv2.VideoCapture(args.i)
    cap.set(cv2.CAP_PROP_POS_FRAMES, start)

    def animate(video_frame_in_this_thread):
        """
        This animates the FuncAnimation
        """
        global_frame = video_frame_in_this_thread + start

        if args.z:
            axs.set(
                xlim=(x_min[global_frame], x_max[global_frame]),
                ylim=(y_max[global_frame], y_min[global_frame]),
            )

        points.set_offsets(pos[global_frame])
        im.set_data(np.mean(cap.read()[1], axis=2))

        origin = max(0, global_frame - line_lenght)
        for fish in range(N):
            LineCollections[fish].set_segments(segments[origin:global_frame, fish])
        return

    ani = FuncAnimation(
        fig,
        animate,
        frames=trange(
            n_frames,
            total=n_frames - 1,
            position=thread,
            desc=f"Thread {thread}",
        ),
        init_func=lambda: 0,
    )

    filename = f"{tmp_folder}part{thread:02d}.mp4"
    ani.save(filename, fps=fps)
    return filename


rmtree(tmp_folder, ignore_errors=True)
makedirs(tmp_folder)


def main():
    with Pool(args.n) as p, open("files.txt", "w") as file:
        for filename in p.map(thread_run, range(args.n)):
            file.write("file " + filename + "\n")
    print("\n" * args.n)

    call(
        [
            "ffmpeg",
            "-safe",
            "0",
            "-hide_banner",
            "-loglevel",
            "error",
            "-f",
            "concat",
            "-y",
            "-i",
            "files.txt",
            "-c",
            "copy",
            args.o if args.o else splitext(args.i)[0] + "_tracked.mp4",
        ]
    )

    rmtree(tmp_folder)
    remove("files.txt")


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import numpy as np
import cv2
from rich import print
from functools import lru_cache
import os
from rich.console import Console
from id_manual_tools.matplotlib_gui import matplotlib_gui
from rich.table import Table
from id_manual_tools.utils import trajectory_path
from argparse import ArgumentParser
from rich.align import Align


console = Console()


class setup_points_setter(matplotlib_gui):
    def __init__(self, video_path, data, name):

        self.cap = cv2.VideoCapture(video_path)
        self.total_frames = self.cap.get(cv2.CAP_PROP_FRAME_COUNT)
        print(f"Loaded {os.path.abspath(video_path)}")

        self.data = data

        self.setup_points = {}
        for name_i, points in self.data["setup_points"].items():
            self.setup_points[name_i] = [list(points[:, 0]), list(points[:, 1])]

        self.name = name

        if not self.name in self.setup_points:
            self.setup_points[self.name] = [[], []]
            console.rule(f"[bold red]Creating setup points: {self.name}")
        else:
            console.rule(f"[bold red]Modifying setup points: {self.name}")

        self.xmax = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.ymax = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

        self.create_figure()
        self.actual_plotted_frame = -1
        self.frame = 0
        self.x_center = self.xmax / 2
        self.y_center = self.ymax / 2

        self.zoom = self.ymax / (2 * self.Ly)
        self.set_ax_lims(do_not_draw=True)

        self.draw_frame()
        plt.show()

    def draw_frame(self):

        if self.frame != self.actual_plotted_frame:
            self.im.set_data(self.get_frame(self.frame))

            self.text.set_text(f"Frame {self.frame}")
            self.actual_plotted_frame = self.frame

        self.draw_and_flush()

    def create_figure(self):

        super().__init__("Setup points editor")

        self.im = self.ax.imshow(
            [[[0, 0, 0]]],
            extent=(
                0,
                self.xmax,
                self.ymax,
                0,
            ),
            interpolation="none",
            animated=True,
        )

        self.lines = {
            name: self.ax.plot(*self.close_line(*xy), ":.", label=name)[0]
            for name, xy in self.setup_points.items()
        }
        self.ax.legend()

        self.text = self.ax.text(
            0.1, 0.1, "", size=15, transform=self.ax.transAxes, zorder=15
        )

    @staticmethod
    def close_line(x, y):
        if len(x) < 3:
            return x, y
        else:
            return x + [x[0]], y + [y[0]]

    def key_a(self):
        self.frame = max(0, self.frame - 1)
        self.draw_frame()

    def key_d(self):
        self.frame = min(self.total_frames - 1, self.frame + 1)
        self.draw_frame()

    def key_left(self):
        self.key_a()

    def key_right(self):
        self.key_d()

    def key_enter(self):
        out_dict = {}
        for name, points in self.setup_points.items():
            out_dict[name] = np.array(points).astype(int).T
        self.data["setup_points"] = out_dict
        print(f"[green]Data writed in trajectory file")
        plt.close()

    @lru_cache(maxsize=20)
    def get_frame(self, frame):

        if self.cap.get(cv2.CAP_PROP_POS_FRAMES) != frame:
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
        ret, image = self.cap.read()
        assert ret

        return image

    @staticmethod
    def closest_point(x, y, xs, ys):
        distances = [(x - x_i) ** 2 + (y - y_i) ** 2 for x_i, y_i in zip(xs, ys)]
        return np.argmin(distances), np.min(distances)

    @staticmethod
    def sort_points(x, y):

        atan2 = np.arctan2(y - np.mean(y), x - np.mean(x))
        x = [
            (x_i[1], x_i[2])
            for x_i in sorted(zip(atan2, x, y), key=lambda pair: pair[0])
        ]
        return list(map(list, zip(*x)))

    def button_1(self, event):
        self.setup_points[self.name][0].append(event.xdata)
        self.setup_points[self.name][1].append(event.ydata)
        self.setup_points[self.name] = self.sort_points(*self.setup_points[self.name])

        self.lines[self.name].set_data(*self.close_line(*self.setup_points[self.name]))
        self.draw_frame()

    def button_3(self, event):

        point_id, dist = self.closest_point(
            event.xdata, event.ydata, *self.setup_points[self.name]
        )
        if dist < 1000:
            self.setup_points[self.name][0].pop(point_id)
            self.setup_points[self.name][1].pop(point_id)
            self.lines[self.name].set_data(
                *self.close_line(*self.setup_points[self.name])
            )
            self.draw_frame()


def rename_setup_point(setup_points):
    old_name = console.input(f"Enter the OLD name of the setup points: ")
    if old_name in setup_points:
        new_name = console.input(f'Enter the NEW name for setup points "{old_name}": ')
        setup_points[new_name] = setup_points.pop(old_name)
        console.rule(f'[green]Succesfully renamed points "{old_name}" to "{new_name}"')
    else:
        print(f'[red]No setup points named "{old_name}", aborting...')


def delete_setup_point(setup_points):
    name = console.input(f"Enter the name of the setup points to delete: ")
    if name in setup_points:
        setup_points.pop(name)
        console.rule(f'[green]Succesfully deleted "{name}"')
    else:
        print(f'[red]No setup points named "{name}", aborting...')


def setup_points_table(setup_points):

    setup_points = [(name, len(points)) for name, points in setup_points.items()]
    table = Table(title="setting_points")

    table.add_column("key", justify="center", style="cyan")
    table.add_column("# of points", justify="center", style="magenta")

    for key, length in setup_points:
        table.add_row(key, f"{length}")

    return table


def arg_main(video_path, data):
    console.rule("Welcome to the setup_points setter!")

    if isinstance(data, str):
        input_arg = "path"
        path = trajectory_path(data)
        data = np.load(path, allow_pickle=True).item()
    else:
        input_arg = "data"

    if not isinstance(data["setup_points"], dict):
        data["setup_points"] = {}
        print(f"[green]Created an empty dict for setup_points")

    while True:
        console.print(Align.center(setup_points_table(data["setup_points"])))
        what_to_do = console.input(
            "[blue]Name an existing (or not) setup_points to modify it or deletet it (or create it). You can also write 'rename' or 'delete' to rename or delete an existing setup_point or 'q' to save and exit the app: "
        )
        if what_to_do == "q":
            if input_arg == "path":
                np.save(path, data)
                console.print(f"Data saved to {path}")
            console.rule("bye!")
            return
        elif what_to_do == "rename":
            rename_setup_point(data["setup_points"])
        elif what_to_do == "delete":
            delete_setup_point(data["setup_points"])
        elif what_to_do:
            setup_points_setter(video_path, data, name=what_to_do)


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "-s",
        metavar="session",
        type=trajectory_path,
        help="idTracker.ai succesful session directory or trajectory file",
        required=True,
    )
    parser.add_argument(
        "-video",
        type=str,
        help="Video file (only one file)",
        required=True,
    )
    args = parser.parse_args()
    arg_main(args.video, args.s)


if __name__ == "__main__":
    main()



import numpy as np
import csv
from id_manual_tools.utils import trajectory_path


def find_nans_1D(data, id=None):
    """Returns a list of the nans location for a 1D array:

        (fish_id (optional), start, end, length of slice)

    In such a way that data[start] is the first nan and data[end-1] is the last nan
    """
    assert (
        data.ndim == 1
    ), f"Only one dimensional arrays, the given array has shape {data.shape}"

    dif = np.diff(np.concatenate(([-1], np.where(~np.isnan(data))[0], [len(data)])))
    end = np.cumsum(dif) - 1
    nan = dif - 1
    valid = nan > 0
    if id is not None:
        return [(id, e - n, e, n) for e, n in zip(end[valid], nan[valid])]
    else:
        return [(e - n, e, n) for e, n in zip(end[valid], nan[valid])]


def get_list_of_nans_from_traj(traj, sort_by="length"):
    """Returns a list of the nans location for a trajectory 2D/3D array:

        (fish_id (only for 3D arrays), start, end, length of slice)

    In such a way that traj[fish_id, start] is the first nan and
    traj[fish_id, end-1] is the last nan.
    Elements in the list are sorted by length of slice (reverse)
    """
    if traj.ndim == 3:
        traj = traj[..., 0]

    if traj.ndim == 1:
        nans = find_nans_1D(traj)
    else:
        nans = []
        for fish_id in range(traj.shape[1]):
            nans += find_nans_1D(traj[:, fish_id], fish_id)

    if sort_by == "length":
        nans.sort(key=lambda x: x[-1], reverse=True)
    elif sort_by == "end":
        nans.sort(key=lambda x: x[-2], reverse=True)
    elif sort_by == "start":
        nans.sort(key=lambda x: x[-3], reverse=True)
    elif sort_by == "id":
        nans.sort(key=lambda x: x[-4], reverse=True)
    return nans


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Nans observation in idtrackeri.ai trajectories."
    )
    parser.add_argument(
        "-i",
        type=str,
        help="idTracker.ai succesful session directory or trajectory file",
    )
    parser.add_argument(
        "-o", type=str, default=None, help="output file, default input[:4]+'_nans.csv'"
    )
    args = parser.parse_args()

    input_path = trajectory_path(args.i, read_only=True)

    output_path = args.o if args.o else input_path[:-4] + "_nans.csv"

    traj = np.load(input_path, allow_pickle=True).item()["trajectories"][..., 0]

    nans = get_list_of_nans_from_traj(traj)

    with open(output_path, "w", newline="") as csvfile:
        csvfile.write("fish_id,start,end,duration\n")
        writer = csv.writer(csvfile)
        writer.writerows(nans)
    print(f"File saved at {output_path}")


if __name__ == "__main__":
    main()
